SuperStrict

Rem
bbdoc: Asynchronus Tasks Extended
about: This module contains some simple executor and task types to use in a non-threaded or threaded application.
End Rem
Module fscom.AsyncEx

ModuleInfo "Version: 1.0"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

Import fscom.Async
Import brl.reflection
?threaded
	Import brl.threads
?

Rem 
bbdoc: This executor will execute all tasks immeadiatly in the thread they are submitted.
about: If you don't want to completely rewrite your code for ?threaded and ?Not threaded mode,
you can use this executor to use TTask and TFuture with non-threaded code. It is also useful
to execute a list of tasks in a specific thread (e.g. graphics tasks in the main-thread).
End Rem
Type TEagerSingleThreadedExecutor Extends TExecutor
	Field shutdownState:Byte = False
	?threaded
		Field shutdownMutex:TMutex
		
		Method New()
			shutdownMutex = TMutex.Create()
			shutdownMutex.Lock()
		End Method
	?
	
	Method Submit:TFuture(task:TTask)
		If shutdownState Throw "Executer shut down"
		
		Local future:TEagerSingleThreadedFuture = New TEagerSingleThreadedFuture
		future.task = task
		Try
			future.retVal = task.Execute()
		Catch ex:Object
			DebugLog task.ToString()+": "+ex.ToString()
			future.exception = ex
		End Try
		Return future
	End Method

	Method ExecuteAll:TFuture[](tasks:TTask[] , timeout:Int = TExecutor.TIMEOUT_FOREVER)
		Return SubmitAll(tasks)
	End Method
	
	Method ExecuteAny:Object(tasks:TTask[] , timeout:Int = TExecutor.TIMEOUT_FOREVER)
		If shutdownState Throw "Executer shut down"

		Local time:Int = MilliSecs()
		For Local i:Int = 0 Until tasks.length
			Try
				Return tasks[i].Execute()
			Catch ex:Object
				If timeout >= 0 And time + timeout < MilliSecs()
					Throw "Timeout"
				EndIf
			End Try 
		Next
		Throw "No task returned successfully"	
	End Method
	
	Method Shutdown()
		shutdownState = True
		?threaded
			shutdownMutex.Unlock()
		?
	End Method
		
	Method ShutdownNow()
		Shutdown()
	End Method

	Method WaitShutdown(timeout:Int = TExecutor.TIMEOUT_FOREVER)
		If shutdownState
			Return
		Else
			?threaded
				If timeout < 0
					'Wait for the mutex
					shutdownMutex.Lock()
					shutdownMutex.Unlock()
				Else
					Local time:Int=MilliSecs()
					Repeat
						Delay 10
						If time + timeout < MilliSecs()
							Throw "Timeout"
						EndIf
					Until shutdownMutex.TryLock()
					shutdownMutex.Unlock()
				EndIf
			?Not threaded
				If timeout < 0 'this would result in a deadlock.
					'if there is only one thread and it didn't order a shutdown
					'no amount of time will be enough to wait for other threads
					'to call Shutdown.
					Throw "Timeout"
				Else
					'mimic a waiting for something
					Delay timeout
					Throw "Timeout"
				EndIf
			?
		EndIf
	End Method
	
	Method IsShutdownStarted:Byte()
		Return shutdownState
	End Method
	
	Method IsShutdownComplete:Byte()
		Return shutdownState
	End Method
End Type

Private

'Future returned by TEagerSingleThreadedExecutor
Type TEagerSingleThreadedFuture Extends TFuture
	Field retVal:Object
	Field exception:Object
	Field task:TTask
	
	Method GetTask:TTask()
		Return task
	End Method
	
	Method IsDone:Byte()
		Return True
	End Method
	
	Method IsCanceled:Byte()
		Return False
	End Method
	
	Method Cancel:Byte()
		Return False
	End Method
	
	Method Wait:Object(timeout:Int)
		If exception Then Throw exception
		Return retVal
	End Method
End Type

Public 

Rem 
bbdoc: This executor will execute tasks only if someone calls the Wait method of their TFuture.
about: This executor can be used to "collect" tasks from a piece of code 
that submits tasks to a given executor and call / cancel them later.
End Rem
Type TLazySingleThreadedExecutor Extends TExecutor
	Field shutdownState:Byte = False
	Field pending:TList=New TList
	?threaded
		Field shutdownMutex:TMutex
		Field pendingMutex:TMutex = TMutex.Create()
		
		Method New()
			shutdownMutex = TMutex.Create()
			shutdownMutex.Lock()
		End Method
	?
	
	'Thread-safe removing from the pending list
	Method RemovePending(o:Object)
		?threaded
			pendingMutex.Lock()
		?
		pending.Remove(o)
		?threaded
			pendingMutex.Unlock()
		?
	End Method
	
	'Thread-safe getting all unfinished work
	Method GetPending:TFuture[]()
		?threaded
			pendingMutex.Lock()
		?
		Local arr:TFuture[]=TFuture[](pending.ToArray())
		?threaded
			pendingMutex.Unlock()
		?
		Return arr
	End Method
	
	Method Submit:TFuture(task:TTask)
		If shutdownState Throw "Executer shut down"
		
		Local future:TLazySingleThreadedFuture = New TLazySingleThreadedFuture
		future.task = task
		future.executor = Self
		?threaded
			pendingMutex.Lock()
		?
		pending.AddLast future
		?threaded
			pendingMutex.Unlock()
		?
		Return future
	End Method	

	Method ExecuteAll:TFuture[](tasks:TTask[] , timeout:Int = TExecutor.TIMEOUT_FOREVER)
		Local futures:TFuture[] = SubmitAll(tasks)
		Local time:Int=MilliSecs()
		For Local i:Int = 0 Until futures.length
			If timeout<0
				futures[i].Wait()
			Else	
				futures[i].Wait(timeout - (MilliSecs() - time) )
			EndIf
		Next
		Return futures
	End Method
	
	Method ExecuteAny:Object(tasks:TTask[] , timeout:Int = TExecutor.TIMEOUT_FOREVER)
		If shutdownState Throw "Executer shut down"

		Local time:Int = MilliSecs()
		For Local i:Int = 0 Until tasks.length
			Try
				Return tasks[i].Execute()
			Catch ex:Object
				If timeout >= 0 And time + timeout < MilliSecs()
					Throw "Timeout"
				EndIf
			End Try 
		Next
		Throw "No task returned successfully"	
	End Method
	
	Method Shutdown()
		shutdownState = True
		?threaded
			shutdownMutex.Unlock()
		?
	End Method
		
	Method ShutdownNow()
		For Local future:TFuture = EachIn pending
			future.Cancel()
		Next
		Shutdown()
	End Method

	Method WaitShutdown(timeout:Int = TExecutor.TIMEOUT_FOREVER)
		If shutdownState
			Return
		Else
			?threaded
				If timeout < 0
					'Wait for the mutex
					shutdownMutex.Lock()
					shutdownMutex.Unlock()
				Else
					Local time:Int=MilliSecs()
					Repeat
						Delay 10
						If time + timeout < MilliSecs()
							Throw "Timeout"
						EndIf
					Until shutdownMutex.TryLock()
					shutdownMutex.Unlock()
				EndIf
			?Not threaded
				If timeout < 0 'this would result in a deadlock.
					'if there is only one thread and it didn't order a shutdown
					'no amount of time will be enough to wait for other threads
					'to call Shutdown.
					Throw "Timeout"
				Else
					'mimic a waiting for something
					Delay timeout
					Throw "Timeout"
				EndIf
			?
		EndIf
	End Method
	
	Method IsShutdownStarted:Byte()
		Return shutdownState
	End Method
	
	Method IsShutdownComplete:Byte()
		Return shutdownState
	End Method
End Type

Private

?Not threaded
	Type TLazySingleThreadedFuture Extends TFuture
		Const STATE_NOT_DONE:Byte = 0 , STATE_EXECUTED:Byte = 1 , STATE_CANCELED:Byte = 2
		
		Field task:TTask
		Field retVal:Object
		Field exception:Object
		Field executor:TLazySingleThreadedExecutor
		Field state:Byte=STATE_NOT_DONE
		
		Method GetTask:TTask()
			Return task
		End Method
		
		Method IsDone:Byte()
			Return state <> STATE_NOT_DONE
		End Method
		
		Method IsCanceled:Byte()
			Return state = STATE_CANCELED
		End Method
		
		Method Cancel:Byte()
			If state=STATE_NOT_DONE
				state = STATE_CANCELED
				executor.RemovePending Self
				Return True
			Else
				Return False
			EndIf
		End Method
		
		Method Wait:Object(timeout:Int)
			Select state
			Case STATE_NOT_DONE
				Try
					retVal = task.Execute()
					state = STATE_EXECUTED
					executor.RemovePending Self
					Return retVal
				Catch ex:Object
					exception = ex
					executor.RemovePending Self
					state=STATE_EXECUTED
					Throw ex
				End Try
			Case STATE_EXECUTED
				If exception Throw exception
				Return retVal
			Case STATE_CANCELED
				Throw "Canceled"
			End Select
		End Method
	End Type
?threaded
	Type TLazySingleThreadedFuture Extends TFuture
		Const STATE_IDLE:Int = 0 , STATE_STARTED:Int = 1 , STATE_EXECUTED:Int = 2 , STATE_CANCELED:Int = 3
		
		'use this mutex to sync retval and exception
		Field mutex:TMutex=TMutex.Create()
		Field retVal:Object
		Field executor:TLazySingleThreadedExecutor
		Field exception:Object
		
		'use ony CompareAndSwap to sync state
		Field state:Int = STATE_IDLE

		'read-only
		Field task:TTask
		
		Method GetTask:TTask()
			Return task
		End Method
		
		Method isDone:Byte()
			Local isExecuted:Byte = CompareAndSwap(state , STATE_EXECUTED , STATE_EXECUTED)
			Local isCanceled:Byte = CompareAndSwap(state , STATE_CANCELED , STATE_CANCELED)
			Return Not (isExecuted Or isCanceled)
		End Method
		
		Method isCanceled:Byte()
			Return CompareAndSwap(state , STATE_CANCELED , STATE_CANCELED)
		End Method
		
		Method Cancel:Byte()
			Local isCanceled:Byte = CompareAndSwap(state , STATE_IDLE , STATE_CANCELED)
			If isCanceled
				executor.RemovePending Self
				Return True
			Else
				Return False
			EndIf
		End Method
		
		Method Wait:Object(timeout:Int)
			Repeat
				If CompareAndSwap(state , STATE_EXECUTED , STATE_EXECUTED)..
					Or CompareAndSwap(state , STATE_STARTED, STATE_STARTED)
					'Waits for a started execution to end.
					'The mutex MUST be locked BEFORE state is
					'set To STATE_STARTED for this to work
					mutex.Lock()
					Local retVal_:Object = retVal
					Local exception_:Object = exception
					mutex.Unlock()
					If exception_ Throw exception
					Return retval_
				ElseIf CompareAndSwap(state , STATE_CANCELED , STATE_CANCELED)
					mutex.Unlock()
					Throw "Canceled"
				ElseIf CompareAndSwap(state , STATE_IDLE , STATE_IDLE)
					'First lock mutex before announce that work is started; see above
					If mutex.TryLock()
						If CompareAndSwap(state , STATE_IDLE , STATE_STARTED)
							Try
								Local retVal_:Object = task.Execute()
								retVal = retVal_
								
								'This MUST NOT fail
								CompareAndSwap(state , STATE_STARTED , STATE_EXECUTED)
								
								executor.RemovePending Self
								mutex.Unlock()
								Return retVal_
							Catch ex:Object
								exception = ex

								'This MUST NOT fail
								CompareAndSwap(state , STATE_STARTED , STATE_EXECUTED)

								executor.RemovePending Self
								mutex.Unlock()
								Throw ex
							End Try 
						Else
							'Some other thread already changed the state.
							'since this thread has the mutex locked,
							'it can only be Cancel(). Anyways:
							'Unlock mutex and re-run state evaluation
							mutex.Unlock()
						EndIf
					Else
						'Some other thread was faster with locking.
						're-run state evaluation
					EndIf
				EndIf
			Forever
			'Due unfortunate timing, the value of state could have changed between
			'two calls of CompareAndSwap so no if/else-case will be matched by the first run
		End Method
	End Type
?

Public


Rem
bbdoc: Task to execute a function without parameters
End Rem
Type TSimpleFunctionTask Extends TTask
	Function Create:TTask(func:Object() )
		Local task:TSimpleFunctionTask = New TSimpleFunctionTask
		task.func = func
		Return task
	End Function
	
	Field func:Object()
	
	Method Execute:Object()
		Return func()
	End Method
End Type

Rem
bbdoc: Task to execute a function designed like hook functions of BRL.Hook 
End Rem
Type THookTask Extends TTask
	Function Create:TTask(func:Object(id:Int , data:Object , context:Object) , id:Int , data:Object , context:Object=Null)
		Local task:THookTask = New THookTask
		task.func = func
		task.id = id
		task.data = data
		task.context = context
		Return task
	End Function
	
	Field func:Object(id:Int , data:Object , context:Object)
	Field id:Int
	Field data:Object
	Field context:Object
	
	Method Execute:Object()
		Return func(id , data , context)
	End Method
End Type

Rem
bbdoc: Task to call a method of an object with the given parameters
End Rem
Type TMethodTask Extends TTask
	Function Create:TTask(this:Object , methodName:String , params:Object[])
		Local typeid:TTypeId=TTypeId.ForObject(this)
		Return CreateR(this , typeid.FindMethod(methodName) , params)
	End Function

	Function CreateR:TTask(this:Object , method_:TMethod , params:Object[])
		Local task:TMethodTask = New TMethodTask
		task.this = this
		task.method_ = method_
		If task.method_ = Null Throw "Method not found"
		task.params = params
		Return task
	End Function
	
	Field this:Object
	Field method_:TMethod
	Field params:Object[]
	
	Method Execute:Object()
		Return method_.Invoke(this , params)
	End Method
	
	Method ToString:String()
		Local msg:String = "Task: Invoke "+this.ToString()+":"+method_._selfTypeId.Name()+ "." + method_.Name() + "("
		If params.length > 0
			msg:+params[0].ToString()
			For Local i:Int = 1 Until params.length
				msg:+","+params[i].ToString()
			Next
		EndIf
		msg:+")"
		Return msg
	End Method
End Type