SuperStrict
Rem
bbdoc: Bufferedstream - A brl.socketstream replacement
End Rem
Module fscom.bufferedstream

ModuleInfo "Version: 0.3"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

ModuleInfo "History: 0.3"
ModuleInfo "History: TODO Fixed Read"
ModuleInfo "History: 0.2"
ModuleInfo "History: Changed Eof so if the buffer is not empty, it will always return false."

Import brl.socket
Import brl.stream

Rem
bbdoc: This is the TBufferedStream type.
about: Connect a socket to a buffered stream with one of the three constuctors. A buffered stream has a receive buffer and a send buffer.

#Recv will fill the receive buffer with data from the socket if available,
#Send will send the contents of the send buffer and will empty it. 

Use the methods of TStream to read from the receive buffer and write to the send buffer.

Use the #RecvFrom and #SendTo methods to manage a UDP server socket.
End Rem
Type TBufferedStream Extends TStream	
	Rem
	bbdoc: Close the socket if the stream is closed
	End Rem
	Const FLAG_AUTO_CLOSE:Byte = %1
	
	Rem
	bbdoc: Call #Send if #Flush is called
	End Rem
	Const FLAG_SEND_ON_FLUSH:Byte = %10
	
	Rem
	bbdoc: Call #Recv if the buffer is empty and any read-action should be performed or #Eof is called 
	End Rem
	Const FLAG_RECV_ON_EOF:Byte=%100

	Rem
	bbdoc: Socket Constructor
	about: Wraps a TBufferedStream around an existing socket.
	@flags can be a combination of #FLAG_AUTO_CLOSE, #FLAG_SEND_ON_FLUSH and #FLAG_RECV_ON_EOF
	Combination of all flags allowes you to use the TBufferedStream as a TStream without 
	using the underlying socket or any methods defined by TBufferedStream
	End Rem
	Function Create:TBufferedStream(socket:TSocket,flags:Byte=FLAG_AUTO_CLOSE|FLAG_SEND_ON_FLUSH|FLAG_RECV_ON_EOF)
		Local this:TBufferedStream =New TBufferedStream
		this._socket = socket
		this._flags=flags
		Return this	
	End Function

	Rem
	bbdoc: TCP Client Constructor
	about: Will create a TCP socket and connect it to the given host and port, then calls #Create with the given flags.
	It's not recommended to switch off the #FLAG_AUTO_CLOSE with this method.
	End Rem
	Function CreateTCPClient:TBufferedStream( remoteHost:String,remotePort:Int,flags:Byte=FLAG_AUTO_CLOSE|FLAG_SEND_ON_FLUSH|FLAG_RECV_ON_EOF )
		Local remoteIp:Int=HostIp( remoteHost )
		If Not remoteIp Return Null
		
		Local socket:TSocket=TSocket.CreateTCP()
		If socket
			If socket.Connect( remoteIp,remotePort ) 
				Return Create( socket , flags)
			EndIf
			socket.Close
		EndIf
	End Function

	Rem
	bbdoc: UDP Client Constructor
	about: Will create a UDP socket and connect it to the given host and port, then calls #Create with the given flags.
	It's not recommended to switch off the FLAG_AUTO_CLOSE with this method.
	End Rem
	Function CreateUDPClient:TBufferedStream( remoteHost:String,remotePort:Int,flags:Byte=FLAG_AUTO_CLOSE|FLAG_SEND_ON_FLUSH|FLAG_RECV_ON_EOF)
		Local remoteIp:Int=HostIp( remoteHost )
		If Not remoteIp Return Null
		
		Local socket:TSocket=TSocket.CreateTCP()
		If socket
			If socket.Connect( remoteIp,remotePort ) 
				Return Create( socket , flags )
			EndIf
			socket.Close
		EndIf
	End Function
	
	Field _sendArray:Byte[]=New Byte[0]
	Field _recvArray:Byte[]=New Byte[0]
	Field _socket:TSocket
	Field _flags:Byte
	Field _timeout:Int=60*60*1000	

	Rem
	bbdoc: Receives all available data.
	returns: the number of bytes read
	about: This method will check the sockets #ReadAvail and tries to read everything from the socket.
	End Rem
	Method Recv:Int()
		Local ip:Int , port:Int
		Local bytes:Int = 0, newbytes:Int=0
		Repeat
			newbytes = RecvFrom(ip , port)
			bytes:+ newbytes
		Until newbytes = 0
		Return bytes
	End Method
	
	Rem
	bbdoc: Receives available data from a remote host.
	returns: the number of bytes read.
	about: This method will check the sockets ReadAvail and tries to read everything from the socket.
		However it will only read messages from one host and must be called again to read a message from another host.
		The host the message is from can be retrieved by the ip and port variables.
		This method is useful for UDP-Server-Streams.
	End Rem
	Method RecvFrom:Int(ip:Int Var , port:Int Var)
		Local count:Int = _socket.ReadAvail()
		
		If count = 0
			ip = 0
			port = 0
			Return 0
		End If
		
		Local buf:Byte[]=New Byte[count]

		count = recvfrom_(_socket._socket , buf , count , 0 , ip , port)
		If count<0
			Throw New TSocketException
		ElseIf count=0
			Return 0
		EndIf

		_recvArray=_recvArray[.._recvArray.length+count]
		For Local i:Int=_recvArray.length-count To _recvArray.length-1
			_recvArray[i]=buf[i-_recvArray.length+count]
		Next
		
		Return count
	End Method
	
	Rem
	bbdoc: Connects the socket to the given @ip and @port and sends the buffered data
	returns: The number of bytes sent
	about: This method is useful for UDP-Server-Streams.
	End Rem
	Method SendTo:Int(ip:Int , port:Int)
		_socket.Connect(ip,port)
		Return Send()
	End Method
	
	Rem
	bbdoc: Sends the buffered data
	returns: The number of bytes sent
	End Rem
	Method Send:Int()
		Local res:Int=_socket.Send(_sendArray,_sendArray.length)
		_sendArray=New Byte[0]
		Return res
	End Method

	Rem
	bbdoc: Returns the of available bytes from the socket.
	about: Call #Recv or #RecvFrom to read the available data into the internal buffer.
	End Rem
	Method ReadAvail:Int()
		Return _socket.ReadAvail()
	End Method
	
	Rem
	bbdoc: Get stream end of file status
	returns: -1, 0, or 1
	about: The behavior differs slightly from the recommended in BRL.Stream:
	* 0 will be returned if there is something in the receive buffer or the server is connected
	* 1 will be returned if there is nothing more to read and the server is not connected anymore
	* -1 will be returned if there is nothing to read but the server is still connected.

	So classic "While Not stream.Eof()"-loops will only run if there is something to read in the buffer.

	This method will call #Recv if #FLAG_RECV_ON_EOF is set and the receive buffer is empty
	End Rem
	Method Eof:Int()
		If _recvArray.length > 0
			Return False
		EndIf
		
		If _socket.Connected()
			If _flags & FLAG_RECV_ON_EOF
				Recv()
			EndIf

			If _recvArray.length=0
				Return -True
			Else
				Return False
			EndIf
		Else
			Return True
		EndIf
	End Method
	
	Method Flush:Int()
		If _flags & FLAG_SEND_ON_FLUSH
			Send()
		EndIf 
	End Method
	
	Method Close()
		Super.Close()
		If _flags & FLAG_AUTO_CLOSE
			_socket.Close()
		EndIf
	End Method
	
	Method Read:Int( buf:Byte Ptr , count:Int )
		Local timeout:Int=MilliSecs()+Self._timeout
		While _recvArray.length=0
			Recv()
			If MilliSecs() > timeout
				Throw New TStreamReadException
			EndIf
		Wend

		If count>_recvArray.length
			'TODO: Optimize with MemCopy
			For Local i:Int=0 To _recvArray.length-1
				buf[i]=_recvArray[i]
			Next
			count=_recvArray.length
			_recvArray=New Byte[0]
		Else
			'TODO: Optimize with MemCopy
			For Local i:Int=0 To count-1
				buf[i]=_recvArray[i]
			Next
			_recvArray=_recvArray[count..]
		EndIf
		
		Return count
	End Method
	
	Method Write:Int(buf:Byte Ptr,count:Int)
		Local start:Int=_sendArray.length
		_sendArray = _sendArray[.._sendArray.length + count]
		'TODO: Optimize with MemCopy
		For Local i:Int=0 To count-1
			_sendArray[start+i]=buf[i]
		Next
		Return count
	End Method	
End Type

Type TBufferedStreamFactory Extends TStreamFactory
	Method CreateStream:TBufferedStream( url:Object,proto$,path$,readable:Int,writeable:Int )
		If proto$="tcp" Or proto="udp"
			Local i:Int = path.Find( ":" , 0 ) , server$ , port:Int
			If i>=0
				server = path[..i]
				port = Int(path[i + 1..])
			Else
				server = path
				port = 80
			EndIf
			
			If proto="tcp"
				Return TBufferedStream.CreateTCPClient( server , port)
			Else 
				Return TBufferedStream.CreateUDPClient( server , port)
			EndIf 
		EndIf
	End Method
End Type

New TBufferedStreamFactory
