' lexer.bmx - generic lexer/tokeniser framework

SuperStrict

Import "diff.bmx"
Import "stringtable.bmx"

' Generic token groups
'  Note that the content depends on the used language!
'  A string may not be enclosed in double quotes
'  (like in C/C++: 'a' or <abc>) and operators may be
'  actually identifiers (like in BlitzMax: "And" or "Mod").

Const TOKEN_GROUP_NONE:Int = -1 ' use only as placeholder

Const TOKEN_GROUP_EOF:Int =				$00000001
Const TOKEN_GROUP_EOL:Int =				$00000002

Const TOKEN_GROUP_STRING:Int =			$00000010
Const TOKEN_GROUP_NUMERAL:Int =			$00000020

Const TOKEN_GROUP_OPERATOR:Int =			$00000100
Const TOKEN_GROUP_SEPARATOR:Int =			$00000200

Const TOKEN_GROUP_IDENT:Int =			$00001000
Const TOKEN_GROUP_KNOWN_IDENT:Int =		$00002000
Const TOKEN_GROUP_KEYWORD:Int =			$00004000

Const TOKEN_GROUP_COMMENT:Int =			$00010000

Const TOKEN_GROUP_OTHER:Int =			$80000000

' Reserved Token Flags/Masks
Const TOKEN_FLAG_HIGHLIGHT:Int = 0
Const TOKEN_FLAG_PARSER1:Int = 1
Const TOKEN_FLAG_PARSER2:Int = 2
Const TOKEN_FLAG_PARSER3:Int = 3

' Extended Data
Const TOKEN_EXT_KEYWORD:String = "*keyword" ' a keyword within another token (value = keyword, intval = offset position)
Const TOKEN_EXT_STRING:String = "*string" ' a string within another token (value = string, intval = offset position)
Const TOKEN_EXT_REFERENCE:String = "*reference" ' token contains a (file) reference

' Lexer base type
Type TLexer

	Field Source:String
	Field TokenList:TLexerTokenList
	Field Current:TLexerToken
	Field Flags:Int

	Field Keywords:TStringTable
	Field KnownIdents:TStringTable
	Field NoUpdateSource:Int = False

	' Statistics
	Field LastOp:String
	Field LastOpTime:Int

	Method New()
		TokenList = New TLexerTokenList
		TokenList.Parent = Self
	EndMethod

	Method Name:String() Abstract

	'MARK: Source management

	' SetSource discards the current token list and
	'  rebuilds the token list using the new source.
	Method SetSource(NewSource:String Var) Final
		Local ms1:Int , ms2:Int

		ms1 = MilliSecs()
		TokenList.Free()
		Source = NewSource

		Local Pos:Int = -1
		Local Token:TLexerToken

		Repeat
			Token = FindNextToken(Pos)
			If Token Then
				TokenList.Append Token
'				Print Token.ToString()

			Else
				Exit
			EndIf
		Until Token.Group = TOKEN_GROUP_EOF

		TidyTokenList()

		ms2 = MilliSecs()
		LastOp = "SetSource"
		LastOpTime = ms2-ms1

	EndMethod

	' AutoUpdateSource calculates the changed part of the source
	'  and calls UpdateSource
	Method AutoUpdateSource(ChangedSource:String Var, MaxFirstValid:Int = -1, MaxLastValid:Int = -1) Final
		Local ms1:Int , ms2:Int

		Local FirstValid:Int = 0
		Local LastValid:Int = 0

		ms1 = MilliSecs()

		If Not DiffString(Source, ChangedSource, FirstValid, LastValid, False)
			' the strings are equal
			If MaxFirstValid <> -1 And MaxLastValid <> -1 Then ' force update source
				ms2 = MilliSecs()
				UpdateSource ChangedSource , MaxFirstValid , MaxLastValid
				LastOp = "AutoUpdateSource -> " + LastOp + "(forced)"
				LastOpTime = (ms2-ms1) + LastOpTime
			EndIf
			Return
		EndIf

		If MaxFirstValid <> -1 Then FirstValid = Min(MaxFirstValid, FirstValid)
		If MaxLastValid <> -1 Then LastValid = Min(MaxLastValid, LastValid)

		ms2 = MilliSecs()

		If FirstValid = 0 And LastValid = 0 Then
			SetSource ChangedSource
		Else
			UpdateSource ChangedSource , FirstValid , LastValid
		EndIf

		LastOp = "AutoUpdateSource -> " + LastOp
		LastOpTime = (ms2-ms1) + LastOpTime
	EndMethod

	' UpdateSource updates the token list by re-tokenising
	'  the changed part of the source.
	'  FirstValid : The number of leading unchanged characters
	'  LastValid : The number of trailing unchanged characters
	Method UpdateSource(ChangedSource:String Var , FirstValid:Int , LastValid:Int) Final
		Local ms1:Int , ms2:Int
		Local LastValidToken:TLexerToken
		Local FirstValidToken:TLexerToken
		Local Temp:TLexerToken, Temp2:TLexerToken
		Local RemTok:Int = 0, AddTok:Int = 0

		If TokenList.IsEmpty() Or NoUpdateSource Then
			SetSource ChangedSource
			LastOp = "UpdateSource(FV:"+FirstValid+",LV:"+LastValid+") -> " + LastOp
			Return
		EndIf

		ms1 = MilliSecs()
		' Find the last token within the valid first part
		Temp = TokenList.First
		While Temp
			If (Temp.LastChar < FirstValid - 1 - 3) Then
				LastValidToken = temp
			Else
				Exit
			EndIf
			Temp = Temp.NextToken()
		Wend

		' Find the first token within the valid last part
		Temp = TokenList.Last
		While Temp
			If (Temp.FirstChar > Source.length - LastValid + 3) Then
				FirstValidToken = temp
			Else
				Exit
			EndIf
			Temp = Temp.PrevToken()
		Wend
		Temp = Null

		' Remove invalid tokens
		If LastValidToken Then
			If FirstValidToken Then
				' Remove between LastValidToken and FirstValidToken
				While LastValidToken.NextToken() And LastValidToken.NextToken() <> FirstValidToken
					LastValidToken.NextToken().Remove()
					RemTok:+ 1
				Wend
				' Adjust token coordinates
				Local Offset:Int = ChangedSource.Length - Source.Length
				Local Temp:TLexerToken = FirstValidToken
				While Temp
					Temp.FirstChar:+ Offset
					Temp.LastChar:+ Offset
					Temp = Temp.NextToken()
				Wend
			Else
				' Remove after LastValidToken
				While LastValidToken.NextToken()
					LastValidToken.NextToken().Remove()
					RemTok:+ 1
				Wend
			EndIf
		Else ' If Not LastValidToken Then
			If FirstValidToken Then
				' Remove before FirstValidToken
				While FirstValidToken.PrevToken()
					FirstValidToken.PrevToken().Remove()
					RemTok:+ 1
				Wend
				' Adjust token coordinates
				Local Offset:Int = ChangedSource.Length - Source.Length
				Local Temp:TLexerToken = FirstValidToken
				While Temp
					Temp.FirstChar:+ Offset
					Temp.LastChar:+ Offset
					Temp = Temp.NextToken()
				Wend
			Else
				' Retokenise whole source
				SetSource ChangedSource
				LastOp = "UpdateSource -> " + LastOp
				Return
			EndIf
		EndIf

		' Set new source
		Source = ChangedSource

		' Calculate boundaries
		Local Pos:Int = - 1 ' start position
		If LastValidToken Then Pos = LastValidToken.LastChar
		Local FirstValidChar:Int = ChangedSource.Length + 1
		If FirstValidToken Then FirstValidChar = FirstValidToken.FirstChar + ChangedSource.Length - Source.Length				 + 1


		' Tokenise new part
		Local Token:TLexerToken
		Local TempList:TLexerTokenList = New TLexerTokenList
		TempList.Parent = Self

		Repeat
			Token = FindNextToken(Pos)
			If Token Then
				If FirstValidToken Then
					If Token.Match(FirstValidToken) Then
						Token = Null
						Exit ' the current token is the start of the valid last part
					Else
						TempList.Append Token
						AddTok:+ 1

						' remove overlapping tokens
						While FirstValidToken And Token.LastChar >= FirstValidToken.FirstChar
							Temp = FirstValidToken.NextToken()
							FirstValidToken.Remove()
							FirstValidToken = Temp
							RemTok:+ 1
						Wend
					EndIf
				Else
					TempList.Append Token
					AddTok:+ 1
				EndIf
			Else
				Exit
			EndIf
		Until Token.Group = TOKEN_GROUP_EOF

		' Insert new part
		If LastValidToken Then
			If FirstValidToken Then
				' Insert between LastValidToken and FirstValidToken
				Temp2 = LastValidToken ' insert after LastValidToken
				While TempList.First
					Temp = TempList.First.Remove()
					TokenList.InsertAfter Temp, Temp2
					Temp2 = Temp ' insert next token after current token
				Wend
				TempList = Null
			Else
				' Insert after LastValidToken
				While TempList.First
					Temp = TempList.First.Remove()
					TokenList.Append Temp
				Wend
				TempList = Null
			EndIf
		Else ' If Not LastValidToken Then
			If FirstValidToken Then
				' Insert before FirstValidToken
				Temp2 = Null ' insert at the beginning
				While TempList.First
					Temp = TempList.First.Remove()
					TokenList.InsertAfter Temp, Temp2
					Temp2 = Temp ' insert next token after current token
				Wend
				TempList = Null
			Else
				' shouldn't happen...
			EndIf
		EndIf

		TidyTokenList()

		ms2 = MilliSecs()
		LastOp = "UpdateSource(FV:"+FirstValid+", LV:"+LastValid+", -:"+RemTok+", +:"+AddTok+")"
		LastOpTime = (ms2 - ms1)
	EndMethod


	'MARK: Lexing

	' FindNextToken is the heart of the lexer which performs the lexical
	'  analysis and the tokenising.
	'  It must use the "Pos" variable to keep track of the current
	'  position within the source code by setting it to the last
	'  character of the previous token and by expecting it to be
	'  set to the character after which it should start the lexing
	'  and tokenising.
	'  It MUST NOT use any other Field or Global variables.
	'  Any operation that needs knowledge of preceeding and succeeding
	'  tokens must be done within the TidyTokenList method.
	'  It MUST NOT leave any parts of the source code untokenised which
	'  _could_ be interpreted as a token (such as comments).
	Method FindNextToken:TLexerToken(Pos:Int Var) Abstract

	'MARK: Token management

	' TidyTokenList can perform some low-level parsing, such as
	'  merging special groups of tokens into a single one
	'  (e.g. "End" + "If" -> "EndIf") or changing a token's ID (type)
	'  depending on the context (e.g. whether a minus "-" is the
	'  subtraction-operator "a - b" or the negation-operator "a * -b"
	'  It MUST NOT delete any tokens without replacing them.
	'  To skip or ignore certain tokens within the token list
	'  override GetFirstToken, PeekNextToken and PeekPrevToken
	'  and use the Flags parameter if necessary.
	Method TidyTokenList()
	EndMethod

	' Return the first available(!) token; reset token pointer (See: TidyTokenList)
	Method GetFirstToken:TLexerToken(Flags:Int = 0)
		Current = TokenList.First
		Return Current
	EndMethod

	' Return the next available(!) token; forward token pointer. (See: TidyTokenList)
	Method GetNextToken:TLexerToken(Flags:Int = 0) Final
		Current = PeekNextToken(Flags)
		Return Current
	EndMethod

	' Return the next available(!) token. (See: TidyTokenList)
	Method PeekNextToken:TLexerToken(Flags:Int = 0)
		If Not Current Then Return TokenList.First
		If Current Then Return Current.NextToken()
		Return Null
	EndMethod

	' Return the previous available(!) token. (See: TidyTokenList)
	Method PeekPrevToken:TLexerToken(Flags:Int = 0)
		If Current Then Return Current.PrevToken()
		Return Null
	EndMethod

	' Return the token at position "Pos", if available (See: TidyTokenList)
	Method GetTokenAt:TLexerToken(Pos:Int, Flags:Int = 0)
		Local Temp:TLexerToken = TokenList.First
		While Temp
			If (Pos >= Temp.FirstChar) And (Pos <= Temp.LastChar) Then
				Return Temp
			ElseIf (Temp.LastChar < Pos) Then
				Return Null
			EndIf
			Temp = Temp.NextToken()
		Wend
		Return Temp
	EndMethod

	Method IsFirstTokenInLine:Int(Flags:Int = 0)
		If Not Current Then Return False
		If Not Current.PrevToken() Then Return True ' is first in file

		Return Current.PrevToken().Group = TOKEN_GROUP_EOL
	EndMethod



	'MARK: Meta data parser
'
'	Method MetaParserInit() Final
'		MetaFound = False
'		MetaName = ""
'		MetaValue = ""
'		MetaTemp = ""
'		MetaLineStart = 0
'		MetaFirstInLine = True
'	EndMethod
'
'	Method MetaParseChar(Token:TLexerToken, Char:Int, Pos:Int) Final
'		If Not MetaTokenList Then Return
'		If Char = CHAR_CR Or Char = CHAR_LF Then
'			MetaTemp = MetaTemp.Trim()
'			If MetaTemp.Find(":") <> - 1 Then
'				Local Name:String = MetaTemp[..MetaTemp.Find(":")].Trim().ToLower()
'				Local Found:Int = False
'				For Local n:Int = 0 Until MetaTokenList.Length
'					If Name = MetaTokenList[n] Then Found = True
'				Next
'				If Name.StartsWith("*") Then Found = True ' is a special meta token
'				If Found Then
'					If MetaFound Then ' store previous meta data
'						Token.AddExtData MetaName, MetaValue
'					EndIf
'					If Not Name.StartsWith("*") Then Token.AddExtData TOKEN_META_METANAME, Name, MetaLineStart
'					MetaName = Name
'					MetaValue = MetaTemp[MetaTemp.Find(":") + 1..]
'					MetaFound = True
'				Else
'					If MetaFound Then ' add to previous meta data
'						MetaValue:+ " " + MetaTemp
'					EndIf
'				EndIf
'			Else
'				If MetaFound Then ' add to previous meta data
'					MetaValue:+ "~n" + MetaTemp
'				EndIf
'			EndIf
'			MetaTemp = ""
'			MetaLineStart = 0
'			MetaFirstInLine = True
'		Else
'			If MetaFirstInLine Then
'				If Char <> $2A Then ' ignore leading *
'					If Char > $20 Then
'						MetaLineStart = Pos
'						MetaFirstInLine = False
'					EndIf
'					MetaTemp:+ Chr(Char)
'				EndIf
'			Else
'				MetaTemp:+ Chr(Char)
'			EndIf
'		EndIf
'	EndMethod
'
'	Method MetaParserRemove(Chars:Int) Final
'		If Not MetaTokenList Then Return
'		MetaTemp = MetaTemp[..MetaTemp.Length - Chars]
'	EndMethod
'
'	Method MetaParserAdd(Chars:String) Final
'		If Not MetaTokenList Then Return
'		MetaTemp:+ Chars
'	EndMethod
'
'	Method MetaParserQuit(Token:TLexerToken) Final
'		If Not MetaTokenList Then Return
'		If MetaTemp = "" And Not MetaFound Then Return
'		MetaTemp = MetaTemp.Trim()
'		If MetaTemp.Find(":") <> - 1 Then
'			Local Name:String = MetaTemp[..MetaTemp.Find(":")].Trim().ToLower()
'			Local Found:Int = False
'			For Local n:Int = 0 Until MetaTokenList.Length
'				If Name = MetaTokenList[n] Then Found = True
'			Next
'			If Name.StartsWith("*") Then Found = True ' is a special meta token
'			If Found Then
'				If MetaFound Then ' store previous meta data
'					Token.AddExtData MetaName, MetaValue
'				EndIf
'				If Not Name.StartsWith("*") Then Token.AddExtData TOKEN_META_METANAME, Name, MetaLineStart
'				Token.AddExtData Name, MetaTemp[MetaTemp.Find(":") + 1..]
'			Else
'				If MetaFound Then ' add to previous meta data
'					MetaValue:+ " " + MetaTemp
'					Token.AddExtData MetaName, MetaValue
'				EndIf
'			EndIf
'		Else
'			If MetaFound Then ' add to previous meta data
'				MetaValue:+ " " + MetaTemp
'				Token.AddExtData MetaName, MetaValue
'			EndIf
'		EndIf
'		MetaFound = False
'		MetaName = ""
'		MetaValue = ""
'		MetaTemp = ""
'	EndMethod
'
'	Function MetaParserParseDoc(Doc:String, Token:TLexerToken) Final
'		Local temp:String
'		Local FirstInLine:Int = True
'		For Local n:Int = 0 Until Doc.Length ' find special tags ("@tag" or "\tag")
'			If FirstInLine And ((Doc[n+1] = $40) Or (Doc[n+1] = $5C)) Then
'				temp = temp.Trim()
'				If (temp[0] = $40) Or (temp[0] = $5C) Then ' is a tag
'					If temp.Contains(" ") And (temp[1..temp.Find(" ")] <> "") Then
'						Token.AddExtData temp[1..temp.Find(" ")], temp[temp.Find(" ")..]
'					Else
'						Token.AddExtData temp[1..], ""
'					EndIf
'				Else
'					If temp Then Token.AddExtData "intro", temp
'				EndIf
'			Else
'				temp:+ Chr Doc[n]
'			EndIf
'			If Doc[n] > $20 Then FirstInLine = False
'			If Doc[n] = CHAR_LF Then FirstInLine = True
'		Next
'		If (temp[0] = $40) Or (temp[0] = $5C) Then ' is a tag
'			If temp.Contains(" ") Then
'				Token.AddExtData temp[1..temp.Find(" ")], temp[temp.Find(" ")..]
'			Else
'				Token.AddExtData temp[1..], ""
'			EndIf
'		Else
'			Token.AddExtData "intro", temp
'		EndIf
'	EndFunction

	'MARK: Editor Helper Methods

	' Find an expression within the current source.
	'  	Returns: True, when a match has been found
	'   MatchStart, MatchLength: coordinates of the found match
	'   Expression: plain text or Regex
	'   ExcludeGroups: exclude token goups from being searched
	'   IncludeGroups: restrict the search to token goups
	'   FindRegex: Expression is a regular expression
	'
	'	POSIX modern regular expressions (greedy):
	'		Most characters match only themselves, the exceptions are:
	'		.		Matches any single character except CR, LF.
	'		[  ]	Matches a single character that is contained within the brackets:
	'				[abc] matches "a" , "b" or "c"
	'				[a-d] matches "a" to "d"
	'				[a-dw-z] matches "a" to "d" or "w" to "z"
	'				[-123] matches "-" , "1" , "2" or "3"
	'		[^  ]	Matches a single character that is NOT contained within the brackets:
	'				[^abc] matches every character except "a" , "b" or "c"
	'		^		Matches the start of a line.
	'		$		Matches the end of a line.
	'		(  )	Define a block (marked subexpression). What the block matched can be recalled later.
	'		\n		Matches the content of the n-th Block, n goes from 1 to 9
	'		*		Matches the last character, group or block zero or more times.
	'				a* matches "a", "aa", "aaa" , ...
	'				[abc]* matches any combination of "a", "b" or "c" : "cab", "bacab", "aaa", ...
	'				/n* matches zero or more iterations of what the n-th block matched
	'				(  )* matches zero or more iterations of the string that the block matched
	'		+		Matches the last character, group or block one or more times.
	'		?		Matches the last character, group or block zero or one times.
	'		{x,y}	Match the last character, group or block at least x and not more than y times.
	'		|		Match either the expression before or the expression after the "|".
	'				abc|xyz matches "abc" or "xyz"
	'				ab(cde|xyz) matches "abcde" or "abxyz"
	'		The following characters have to be escaped if they are meant literally:
	'		"(" , ")" , "[" , "]" , "{" , "}" , "|" , "." , "*" , "+" , "?" , "^" , "$" , "\"
	'				"c:\\myfile.txt" matches "c:\myfile.txt"
	'				"\[+" matches "[", "[[", "[[[", ...
	'				"\[\+" matches "[+"
	Method FindInSource:Int(MatchStart:Int Var, MatchLength:Int Var, Expression:String, StartFrom:Int = 0, ..
	CaseSensitive:Int = False, ExcludeGroups:Int = 0, IncludeGroups:Int = 0, FindRegex:Int = False)

		If ExcludeGroups <> 0 Then Notify "Excluding groups isn't supported yet." ; Return False
		If IncludeGroups <> 0 Then Notify "Including groups isn't supported yet." ; Return False
		If FindRegex Then Notify "Regex matching isn't supported yet." ; Return False

		If Expression = "" Then Return False

		If CaseSensitive Then
			MatchStart = Source.Find(Expression, StartFrom)
			If MatchStart = - 1 Then ' wrap around
				MatchStart = Source.Find(Expression) ' no match
				If MatchStart = - 1 Then Return False
			EndIf
			MatchLength = Expression.length
			Return True
		Else
			MatchStart = Source.ToLower().Find(Expression.ToLower(), StartFrom)
			If MatchStart = - 1 Then ' wrap around
				MatchStart = Source.ToLower().Find(Expression.ToLower())
				If MatchStart = - 1 Then Return False ' no match
			EndIf
			MatchLength = Expression.length
			Return True
		EndIf

	EndMethod

	Method Pos2Line:Int(Pos:Int)
		Local Line:Int = 1, Curr:Int, Prev:Int
		Pos = Min(Pos, Source.Length-1)
		For Local n:Int = 0 Until Pos
			Curr = Source[n]
			If Curr = CHAR_CR Then
				If Prev = CHAR_LF Then ' is part of LFCR
					Curr = 0
				Else
					Line = True
				EndIf
			ElseIf Curr = CHAR_LF Then
				If Prev = CHAR_CR Then ' is part of LFCR
					Curr = 0
				Else
					Line = True
				EndIf
			EndIf
			Prev = Curr
		Next
		Return Line
	EndMethod

	' Others

	Method Cleanup()

		TokenList.Free()

	EndMethod

	' This should return the proper file-extension, when the given source
	'  seems to be another language-type or -subtype (For example when
	'  an XML file contains an "html" tag it should return "html").
	Method AssumedSourceType:String()
		Return ""
	EndMethod

	Method LineAtPos:Int(Pos:Int)
		Local Line:Int = 1, Char:Int, Prev:Int = 0
		For Local n:Int = 0 Until Source.length
			Char = Source[n]
			If Char = CHAR_LF Then
				If Prev = CHAR_CR Then
					Char = 0 ' CR-LF
				Else
					Line:+ 1
				EndIf
			ElseIf Char = CHAR_CR Then
				If Prev = CHAR_LF Then
					Char = 0 ' LF-CR
				Else
					Line:+ 1
				EndIf
			EndIf
			Prev = Char
		Next
		Return Line
	EndMethod

	' Debugging ONLY

	Method ToString:String()
		Return TokenList.ToString()
	EndMethod

	Method Stats:String()
		Local nLine:Int , nStr:Int , nNum:Int
		Local nOp:Int , nSep:Int , nIdent:Int , nKnown:Int , nKey:Int
		Local nComm:Int , nOther:Int, nTotal:Int

		Local Curr:TLexerToken = TokenList.First

		nLine = 1
		While Curr
			Select Curr.Group
				Case TOKEN_GROUP_EOL ; nLine:+ 1

				Case TOKEN_GROUP_STRING ; nStr:+ 1
				Case TOKEN_GROUP_NUMERAL ; nNum:+ 1

				Case TOKEN_GROUP_OPERATOR ; nOp:+ 1
				Case TOKEN_GROUP_SEPARATOR ; nSep:+ 1

				Case TOKEN_GROUP_IDENT ; nIdent:+ 1
				Case TOKEN_GROUP_KNOWN_IDENT ; nKnown:+ 1
				Case TOKEN_GROUP_KEYWORD ; nKey:+ 1

				Case TOKEN_GROUP_COMMENT ; nComm:+ 1

				Case TOKEN_GROUP_OTHER ; nOther:+ 1

				Default ; nOther:+ 1
			EndSelect
			nTotal:+ 1
			Curr = Curr.NextToken()
		Wend

		Local out:String = ..
			"Current Source:~n" + ..
			"Size: " + (Source.length / 1024) + " kB (" + Source.length + " Bytes)~n" + ..
			"Number of lines: " + nLine + "~n" + ..
			"Number of tokens: " + nTotal + "~n" + ..
			"  Literals: " + (nStr + nNum) + " (" + nStr + " Strings, " + nNum + " Numerals)~n" + ..
			"  Identifiers: " + (nIdent + nKnown + nKey) + " (" + nKey + " Keywords, " + nKnown + " Known Identifiers, " + nIdent + " Others)~n" + ..
			"  Special: " + (nOp + nSep) + " (" + nOp + " Operators, " + nSep + " Separators)~n" + ..
			"  Comments: " + nComm + "~n" + ..
			"  Others: " + nOther + "~n" + ..
			"~n" + Name() + ":~n"
			If Keywords Then out:+ "Keywords: " + Keywords.Count() + "~n" Else  out:+ "Keywords: 0~n"
			If KnownIdents Then out:+ "Known Identifiers: " + KnownIdents.Count() + "~n" Else  out:+ "Known Identifiers: 0~n"
			out:+ "Last Operation: " + LastOp + "~n" + ..
			"Last Operation Time: " + LastOpTime + " ms"

		Return out
	EndMethod

EndType

Type TLexerTokenList Final
	Field First:TLexerToken
	Field Last:TLexerToken

	Field Parent:TLexer

	Method IsEmpty:Int()
		Return First = Null
	EndMethod


	Method Count:Int()
		Local n:Int = 0
		Local Temp:TLexerToken = First
		While Temp
			n:+ 1
			Temp = Temp.NextToken()
		Wend
		Return n
	EndMethod


	' Clear the token list.
	Method Free()
		Local Temp:TLexerToken , Nx:TLexerToken
		Temp = First
		First = Null
		Last = Null
		If Parent Then Parent.Current = Null
		While Temp
			Nx = Temp.NextToken()
			Temp.Remove()
			Temp= Nx
		Wend
	EndMethod

	' Append a token to the token list.
	Method Append(Token:TLexerToken)
		If Last Then
			InsertAfter Token, Last
		Else
			First = Token
			Last = Token
			Token.Prec = Null
			Token.Succ = Null
			Token.Parent = Self
		EndIf
	EndMethod

	Method InsertAfter(NewToken:TLexerToken, AfterToken:TLexerToken)
		If Not AfterToken Then
			NewToken.Succ = First
			If NewToken.Succ Then NewToken.Succ.Prec = NewToken Else Last = NewToken
			First = NewToken
			NewToken.Prec = Null
		Else
			NewToken.Succ = AfterToken.Succ
			If NewToken.Succ Then NewToken.Succ.Prec = NewToken
			AfterToken.Succ = NewToken
			NewToken.Prec = AfterToken
		EndIf
		If AfterToken = Last Then Last = NewToken
		NewToken.Parent = Self
	EndMethod

	Method MergeTokens(StartToken:TLexerToken, EndToken:TLexerToken)
'		Print "merge from: " + StartToken.ToString()
'		Print "until: " + EndToken.ToString()

		StartToken.LastChar = EndToken.LastChar
		StartToken.Postfix = EndToken.Postfix
		StartToken.ResetFlags

		Local Temp:TLexerToken = StartToken.Succ
		Local TempSucc:TLexerToken
		Repeat
			If Temp = Null Then
				Throw "TLexerTokenList.MergeTokens: EndToken not found!"
			EndIf

			TempSucc = Temp.Succ

			If Temp.ExtData Then
				If Not StartToken.ExtData Then StartToken.ExtData = CreateList()
				While Not Temp.ExtData.IsEmpty()
					Local md:TLexerExtData = TLexerExtData(Temp.ExtData.RemoveFirst())
					If (md.Name = TOKEN_EXT_KEYWORD) Then
						md.IntValue:+ Temp.FirstChar - StartToken.FirstChar
					EndIf
					StartToken.ExtData.Addlast md
				Wend
			EndIf

			If Temp = EndToken Then
				EndToken.Remove()
				Exit
			Else
				Temp.Remove()
			EndIf

			Temp = TempSucc
		Forever

'		Print "done: " + ToString()
	EndMethod
Rem
	Method Split:TLexerTokenList(FirstToken:TLexerToken)
		Local tl:TLexerTokenList = New TLexerTokenList ' create new tokenlist
		tl.Parent = Parent

		tl.Last = Last
		tl.First = FirstToken
		Last = FirstToken.Prec
		Last.Succ = Null
		tl.First.Prec = Null

		Local temp:TLexerToken = tl.First
		While temp
			temp.Parent = tl
			temp = temp.succ
		Wend

		Return tl
	EndMethod
EndRem
	Method ToString:String()
		Local ret:String = "" , n:Int = 0
		Local Curr:TLexerToken = First
		While Curr
			ret:+ Curr.ToString() + "~n"
			If Curr.ExtData Then
				For Local mi:TLexerExtData = EachIn Curr.ExtData
					ret:+ "   " + mi.Name + " : " + mi.StrValue + " : " + mi.IntValue + "~n"
				Next
			EndIf
			Curr = Curr.NextToken()
			n:+ 1 ; If n = 1000 Then ret:+ "<cut>~n" ; Exit
		Wend
		Return ret
	EndMethod

EndType

Type TLexerToken Final

	Field FirstChar:Int ' first character of token
	Field LastChar:Int ' last character of token
	Field Prefix:Int, Postfix:Int ' number of pre/postfix characters
	Field Group:Int ' token group
	Field ID:Int ' token id
	Field Flags:Int = $0000FFFF ' flags 0-15 are set by default
	Field ExtData:TList = Null

	Method GetValue:String(Complete:Int = False)
		If Parent Then
			If Complete Then
				Return Parent.Parent.Source[FirstChar..LastChar + 1]
			Else
				Return Parent.Parent.Source[FirstChar + Prefix..LastChar - Postfix + 1]
			EndIf
		EndIf
	EndMethod

	Method AddExtData(Name:String, StrValue:String = "", IntValue:Int = 0)
		If Not ExtData Then ExtData = CreateList()
		Local e:TLexerExtData = New TLexerExtData

		e.Name = Name.Trim()
		Local LastWasSpace:Int = False
		For Local n:Int = 0 Until StrValue.Length ' clean metavalue
			If StrValue[n] <= $20 Then
				If Not LastWasSpace Then e.StrValue:+ " "
				LastWasSpace = True
			Else
				e.StrValue:+ Chr(StrValue[n])
				LastWasSpace = False
			EndIf
		Next
		e.StrValue = e.StrValue.Trim()
		e.IntValue = IntValue
		ExtData.AddLast e
	EndMethod

	' List management

	Field Succ:TLexerToken , Prec:TLexerToken
	Field Parent:TLexerTokenList

	Method NextToken:TLexerToken ()
		Return Succ
	End Method

	Method PrevToken:TLexerToken ()
		Return Prec
	End Method

	Method Remove:TLexerToken()
		If Parent And Parent.First = Self Then Parent.First = Succ
		If Parent And Parent.Last = Self Then Parent.Last = Prec
		If Succ Then Succ.Prec = Prec
		If Prec Then Prec.Succ = Succ
		Succ = Null ; Prec = Null
		Parent = Null
		Return Self
	End Method

	Method Match:Int(OtherToken:TLexerToken)
		If Group <> OtherToken.Group Then Return False
		If FirstChar <> OtherToken.FirstChar Then Return False
		If Prefix <> OtherToken.Prefix Then Return False

		' comments can be merged so
'		If Group = TOKEN_GROUP_COMMENT Then Return True

		If LastChar <> OtherToken.LastChar Then Return False
		If Postfix <> OtherToken.Postfix Then Return False
		If Group <> OtherToken.Group Then Return False
		If ID <> OtherToken.ID Then Return False

		Return True
	EndMethod


	' Flags

	Method GetFlag:Int(n:Int)
		Return (Flags & Int(2^n)) <> 0
	EndMethod

	Method SetFlag(n:Int, Value:Int)
		If Value Then
			Flags:| Int(2^n)
		Else
			Flags:& ~Int(2^n)
		EndIf
	EndMethod

	Method ResetFlags()
		Flags:| $0000FFFF
	EndMethod

	' Debugging ONLY

	Method ToString:String()
		Select Group
			Case TOKEN_GROUP_EOF ; Return "end of file (" + ID + "|" + FirstChar + "|" + LastChar + ")"
			Case TOKEN_GROUP_EOL ; Return "end of line (" + ID + "|" + FirstChar + "|" + LastChar + ")"

			Case TOKEN_GROUP_STRING ; Return "string (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()
			Case TOKEN_GROUP_NUMERAL ; Return "numeral (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()

			Case TOKEN_GROUP_OPERATOR ; Return "operator (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()
			Case TOKEN_GROUP_SEPARATOR ; Return "separator (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()

			Case TOKEN_GROUP_IDENT ; Return "ident (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()
			Case TOKEN_GROUP_KNOWN_IDENT ; Return "known ident (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()
			Case TOKEN_GROUP_KEYWORD ; Return "keyword (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()

			Case TOKEN_GROUP_COMMENT ; Return "comment (" + ID + "|" + FirstChar + "|" + LastChar + "):" + GetValue()

			Case TOKEN_GROUP_OTHER ; Return "other (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()

			Default ; Return "<unknown group> (" + ID + "|" + FirstChar + "|" + LastChar + "): " + GetValue()
		EndSelect
	EndMethod
End Type

Type TLexerExtData
	Field Name:String
	Field StrValue:String
	Field IntValue:Int
EndType


' Lexer Factory

Private
	Global LexerFactories:TLexerFactory
Public

Type TLexerFactory
	Field Succ:TLexerFactory

	Method New()
		Succ = LexerFactories
		LexerFactories = Self
	End Method

	Method CreateLexer:TLexer(ext:String) Abstract
EndType

Function CreateLexer:TLexer(ext:String)
	Local Factory:TLexerFactory = LexerFactories

	ext = ext.Trim().ToLower()

	While Factory
		Local lexer:TLexer = Factory.CreateLexer(ext)
		If lexer Return lexer
		Factory = Factory.Succ
	Wend
EndFunction


' Non-Operational Lexer

Type TNOOPLexer Extends TLexer
	Method Name:String()
		Return "Non-Operational Lexer"
	EndMethod

	Method FindNextToken:TLexerToken(Pos:Int Var)


	EndMethod

	Function Create:TNOOPLexer()
'		Print "NOOP Lexer"
		Return New TNOOPLexer
	EndFunction
EndType

Type TNOOPLexerFactory Extends TLexerFactory
	Method CreateLexer:TLexer(ext:String)
		Return TNOOPLexer.Create()
	EndMethod
EndType

New TNOOPLexerFactory



Private

Const CHAR_CR:Int = $0D
Const CHAR_LF:Int = $0A
 