SuperStrict

Rem
bbdoc: RMI / Server
about: This is the RMI server component.

The server component will manage the handling of multiple clients connecting to the process. The clients will be able to invoke
methods from objects registered with RMIRegisterObjectService (see fscom.rmishared) on the server.

To give you further control of the clients, you can register types as service using #RMIRegisterTypeService. A new object will then
be created for every client that connects. You can also access the client's service registry and call methods on the client.
End Rem
Module fscom.rmiserver

ModuleInfo "Version: 0.2"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

ModuleInfo "History: 0.2"
ModuleInfo "History: Added TRMITypeService. These services create one object per client and allow callbacks."

Import brl.timer
Import fscom.bufferedstream
Import fscom.rmishared
Import fscom.proxy

Rem 
bbdoc: Start RMI Server
about: This will start an RMIServer on the given port and 
will create a timer to poll the network with the given hertz.
You'll need to have a WaitSystem/WaitEvent/PollSystem call in 
your main loop or have polling activated. 
end rem 
Function RMIStartServer(port:Short = 55555,executor:TExecutor=Null)
	If socket Throw "RMI already started"
	
	socket = TSocket.CreateTCP()
	
	If Not socket.Bind(port)
		socket.Close()
		Throw "Unable to start RMI server"
	EndIf
	socket.Listen(0)
	
	If executor
		serverExecutor = executor
	EndIf
	
	RMIRegisterObjectService TRMIServiceRegistry.GetRegistry(),"TRMIServiceRegistry"
	
	AddHook RMI_POLL_NETWORK_HOOK , PollNetwork , Null 'TODO: server object for multiple servers
End Function

Rem
bbdoc: Register an type as a RMI service
about: Registering an type on a RMI server will make the server 
creating a new instance for each client.

If the type has a Method OnConnect(client: #TRMIClient), this method will
be called right after an instance was created by RMI.

If the type has a Method OnDisconnect(), it will be called after the
client disconnected and the instance was removed from the RMI service registry.

Use the client object given in the OnConnect method to get client information, the other
service instances associated with the same client and remote services of the client.
End Rem
Function RMIRegisterTypeService(typeName:String , serviceName:String = "")
	Local service:TRMITypeService = New TRMITypeService
	service.typeid = TTypeId.ForName(typeName)
	If service.typeid = Null
		Throw "Type not found"
	ElseIf service.typeid = StringTypeId
		Throw "Strings are not supported as RMI service type"
	ElseIf service.typeid.ExtendsType(ArrayTypeId)
		Throw "Arrays are not supported as RMI service type"
	EndIf 
	
	If serviceName = ""
		serviceName = service.typeid.Name()
	EndIf
	service.name=serviceName

	service.onConnect = service.typeid.FindMethod("OnConnect")
	If (Not service.onConnect) Or (service.onConnect.ArgTypes().length <> 1) Or (service.onConnect.ArgTypes()[0] <> TRMIClientTypeId)
		service.onConnect = Null
	EndIf	
	
	service.onDisconnect = service.typeid.FindMethod("OnDisconnect")
	If (Not service.onDisconnect) Or (service.onDisconnect.ArgTypes().length <> 0)
		service.onDisconnect=Null
	EndIf
	
	service.Register()
	
	For Local c:TRMIClientImpl = EachIn TRMIClientImpl.list
		c.StartService(service)
	Next 
End Function

Rem
bbdoc: Represents an client connected to RMI
End Rem
Type TRMIClient Abstract
	Rem
	bbdoc: Get the ip of the client
	End Rem
	Method GetIp:Int() Abstract
	Rem
	bbdoc: Get the port of the client
	End Rem
	Method GetPort:Short() Abstract
	Rem
	bbdoc: Get the protocol, the client is using. Will be "tcp" TODO: or "udp"
	End Rem
	Method GetProtocol:String() Abstract
	Rem
	bbdoc: Will get the instance of the type service that is associated with the client
	End Rem
	Method GetLocalServiceObject:Object(name:String) Abstract
	Rem
	bbdoc: Will get a proxy object of a client's service. Acts like RMIGetServiceObject (see fscom.rmiclient).
	End Rem
	Method GetRemoteServiceObject:Object(serviceName:String , localProxyClass:String = "") Abstract
	Rem
	bbdoc: Sets a shared object associated with the client (e.g. login information)
	returns: the old shared object
	End Rem
	Method SetSharedObject:Object(obj:Object) Abstract
	Rem
	bbdoc: Gets the shared object
	End Rem
	Method GetSharedObject:Object() Abstract
End Type

Rem
bbdoc: Remove a service type from the service registry
End Rem
Function RMIUnregistertypeService(typeName:String)
	Local service:TRMITypeService = TRMITypeService(TRMIService.GetServiceByName(typeName) )
	If service = Null Then Return
	
	service.Unregister()
End Function


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

'BUG: BLITZMAXBUG: I wanted to put this global in the TRMIClient type but reflection seems to
'be unable to find a type by name during initialization
Global TRMIClientTypeId:TTypeId=TTypeId.ForName("TRMIClient")
Global socket:TSocket
Global serverExecutor:TExecutor = New TEagerSingleThreadedExecutor

Type TRMIClientImpl Extends TRMIClient
	'TODO: Evaluate if a map is neccasary. 
	'Global clientmap:TMap=New TMap
	Global list:TList=New TList
	
	Function Create:TRMIClientImpl(ip:Int , port:Short, ctx:TRMIContext)
		Local this:TRMIClientImpl = New TRMIClientImpl
		this.ip = ip
		this.port = port
		this.ctx = ctx
		this.ctx.serviceSuffix = RMI_PRIVATE_FLAG + this.ToString()
		list.AddLast(this)
		'clientmap.Insert(c , c)
		Return this
	End Function

	Field localServices:TMap=New TMap
	'Field remoteServices:TMap 'TODO: cache objects
	Field ip:Int
	Field port:Short 
	Field ctx:TRMIContext
	Field sharedObject:Object
	
	Method GetIp:Int()
		Return ip
	End Method
	
	Method GetPort:Short()
		Return port
	End Method
	
	Method GetProtocol:String()
		Return "tcp"
	End Method 
	
	Method StartService(service:TRMITypeService)
		Local instance:Object = service.typeid.NewObject()
		If service.onConnect
			service.onConnect.Invoke instance , [Self]
		EndIf
		
		RMIRegisterObjectService instance , service.name + RMI_PRIVATE_FLAG + Self.ToString()
		
		localServices.Insert service.name, service
	End Method
	
	Method StopService(service:TRMITypeService)
		Local name:String=service.name + RMI_PRIVATE_FLAG + Self.ToString()
		Local objectService:TRMIObjectService = TRMIObjectService(TRMIService.GetServiceByName(name))
		
		RMIUnregisterObjectService objectService.GetServiceObject()

		If service.onDisconnect
			service.onDisconnect.Invoke objectService.GetServiceObject(), Null
		EndIf
		
		localServices.Remove service.name
	End Method
	
	Method GetLocalServiceObject:Object(name:String)
		Return TRMIObjectService(localServices.ValueForKey(name)).GetServiceObject()
	End Method
	
	Method GetRemoteServiceObject:Object(serviceName:String, localProxyClass:String="")
		If Not localProxyClass
			localProxyClass = serviceName
		EndIf
		
		'copy ctx into the invoke remote context
		Local ctx_:TRMIInvokeRemoteContext = New TRMIInvokeRemoteContext
		ctx_.stream = ctx.stream
		?threaded
			ctx_.streamMutex = ctx.streamMutex
		?
		ctx_.serviceSuffix = ctx.serviceSuffix
		ctx_.executor = ctx.executor
		
		ctx_.serviceName = serviceName
	
		Return CreateProxy(localProxyClass , RMIInvokeRemote , ctx_)
	End Method
	
	Method SetSharedObject:Object(obj:Object)
		Local ret:Object = sharedObject
		sharedObject = obj
		Return ret
	End Method
	
	Method GetSharedObject:Object()
		Return sharedObject
	End Method
	
	Method Compare:Int(with:Object)
		If TRMIClientImpl(with)
			If ip = TRMIClientImpl(with).ip
				Return port - TRMIClientImpl(with).port
			Else
				Return ip - TRMIClientImpl(with).ip
			EndIf 
		Else
			Return Super.Compare(with)
		EndIf
	End Method
	
	Method ToString:String()
		Return DottedIP(ip)+":"+port
	End Method
End Type

Type TRMITypeService Extends TRMIService
	Global list:TList = New TList
	
	Field name:String
	Field typeid:TTypeId
	Field onConnect:TMethod
	Field onDisconnect:TMethod
	
	Method GetName:String()
		Return name
	End Method
	
	Method GetTypeId:TTypeId()
		Return typeid
	End Method
	
	Method Register()
		Super.Register()
		TRMITypeService.list.AddLast(Self)
	End Method

	Method Unregister()

		Super.Unregister()
		TRMITypeService.list.Remove(Self)
	End Method
End Type

Function PollNetwork:Object(hookid:Int,data:Object,context:Object)
	Repeat
		Local csocket:TSocket = socket.Accept(0)
		If csocket
			Local ctx:TRmiContext = New TRMIContext
			ctx.stream = TBufferedStream.Create(csocket)
			?threaded
				ctx.streamMutex = TMutex.Create()
			?
			ctx.executor = serverExecutor
			Local c:TRMIClientImpl = TRMIClientImpl.Create(csocket.RemoteIP() , csocket.RemotePort(),ctx)
			
			'TODO: add suport  for custom stream wrappers (i.e. SSL/TLS to add a security layer)
			'Maybe like this...
			Rem
			Local userStream:TStream=TStream(RunHooks(RMICustomStreamHook, c.stream))
			
			If userStream
				c.stream = userstream
			EndIf
			End Rem
			
			For Local service:TRMITypeService = EachIn TRMITypeService.list
				c.StartService(service)
			Next
		Else 
			Exit 
		EndIf
	Forever 

	
	Local i:Int=0
	For Local c:TRMIClientImpl = EachIn TRMIClientImpl.list
		If c.ctx.stream.Eof()=True
			TRMIClientImpl.list.Remove(c)
			
			For Local service:TRMITypeService = EachIn TRMITypeService.list
				c.StopService(service)
			Next
			
			Continue 
		EndIf
		
		While (Not c.ctx.stream.Eof()) And (RMI_MAX_INVOCATIONS_AT_ONCE<0 Or i<RMI_MAX_INVOCATIONS_AT_ONCE)
			RMIPollStream(c.ctx)
			i:+1
		Wend
	Next 

	Return data
End Function