SuperStrict

Rem
bbdoc: RMI / Shared
about: These are the shared ressources for RMI.

To register a service object on a client or a server use #RMIRegisterObjectService.

Check your registered services by using #TRMIServiceRegistry's method GetInstance(). 
Use RMIGetServiceObject (see fscom.rmiclient) and TRMIClient.GetRemoteServiceObject (see fscom.rmiserver) with
the service name "TRMIServiceRegistry" to get information about the services you can access on your network partner.

This file implements the fscom.rmi protocol:
<pre>
Every message looks like this:
CHAN CMD DATA

CHAN: a byte representing a channel.
 0-254: rmiblocking
 255: rminonblocking

CMD: a char (ascii byte) representing the command
 i: invoke a method on channel CHAN. The answer will be awaited on CHAN if CHAN is not 255.
 r: return. This will be sent as answer to an invoke (only with CHAN <> 255)
 t: throw. This will be sent as answer to an invoke (only with CHAN <> 255)
 other values are invalid in this version.

DATA: Data is different for every CMD
 CMD = i: 
  - BBString service name
  - BBString method name
  - for each parameter:
     serialized object of the parameter. Primitives are represented by their string representations (like in BRL.Reflection)
  
  There is no metadata sent, so there will be errors thrown if the sent invocation won't match up with the local implementation.
 
 CMD = r: serialized object that was returned.
 CMD = t: serialized object that was thrown.
</pre>


End Rem
Module fscom.rmishared

ModuleInfo "Version: 0.3"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

ModuleInfo "History: 0.3"
ModuleInfo "History: Rewrote API to fix errors with concurrent invocations."
ModuleInfo "History: Added threading support"
ModuleInfo "History: 0.2"
ModuleInfo "History: Changed RMIRegisterServiceObject to RMIRegisterObjectService"
ModuleInfo "History: Moved RMIInvokeRemote and RMIInvokeLocal here both"

Import brl.hook
Import fscom.serialization
Import fscom.asyncex

Rem
bbdoc: Limit the invocations done by the server or a client at once when #RMIPollNetwork is called.
about: Use -1 to not limit the invocations.
End Rem
Global RMI_MAX_INVOCATIONS_AT_ONCE:Int=-1

Rem
bbdoc: Register an object as a RMI service
about: Registering an object will expose the object's methods to be called
by RMI. Use @name to set a name different to the type name of the object.
End Rem
Function RMIRegisterObjectService(serviceObject:Object , name:String = "")
	Local service:TRMIObjectServiceImpl = New TRMIObjectServiceImpl
	service.serviceObject = serviceObject
	service.typeid=TTypeId.ForObject(serviceObject)
	
	If name = ""
		name = service.typeid.Name()
	EndIf
	
	service.name=name
	service.Register()
End Function

Rem
bbdoc: Remove an object from the RMI service registry
End Rem
Function RMIUnregisterObjectService(serviceObject:Object)
	Local service:TRMIObjectServiceImpl = TRMIObjectServiceImpl(TRMIObjectServiceImpl.objectMap.ValueForKey(serviceObject))
	If Not service Then Return 
	
	service.Unregister()
End Function

Rem
bbdoc: Poll the network for new invocations
about: Call this method where it is called periodically e.g. your main loop.
Server and/or client will poll the network and submit invocations inside this method.
End Rem
Function RMIPollNetwork()
	RunHooks RMI_POLL_NETWORK_HOOK, Null
End Function 


Rem
bbdoc: Switch default method blocking on or off. See #RMIInvokeRemote
End Rem
Global RMI_DEFAULT_METHOD_BLOCKING:Byte=True 

Rem
bbdoc: Define timeout for method blocking.
End Rem
Global RMI_BLOCKING_TIMEOUT:Int=60*60*1000

Rem
bbdoc: RMI Exceptions
about: These Exceptions are thrown by RMI. They are unable to be serialized, 
so you can be sure they are thrown by the RMI in your process, not the remote process.
End Rem
Type TRMIException
	Function Create:TRMIException(message:String="",cause:Object=Null )
		Local this:TRMIException = New TRMIException
		this.message = message
		this.cause = cause
		Return this
	End Function
	
	Field message:String
	Field cause:Object
	
	Method ToString:String()
		Local str:String="TRMIException"
		If message
			str:+ ": " + message
		EndIf
		
		If cause
			str:+ " Caused by: " + cause.ToString()
		EndIf
		
		Return str
	End Method
End Type

Rem
bbdoc: Use this flag in the service name to mark it as private. 
about: Private services are not seen in the registry
End Rem
Const RMI_PRIVATE_FLAG:String="@"


Rem
bbdoc: Service Registry
about: This class provides information about the registered services (metadata). 
Call GetRegistry() to get the local registry or get the registry of your network partner by 
requesting the service object for the "TRMIServiceRegistry".
End Rem
Type TRMIServiceRegistry Abstract {proxy}
	Rem
	about: Returns the singleton registry for this process
	End Rem
	Function GetRegistry:TRMIServiceRegistry()
		If Not TRMIServiceRegistryImpl.instance
			TRMIServiceRegistryImpl.instance = New TRMIServiceRegistryImpl
		EndIf
		Return TRMIServiceRegistryImpl.instance
	End Function
	
	Rem
	bbdoc: Get an array of service names
	End Rem
	Method GetServices:String[]() Abstract {rmiblocking}
	
	Rem
	bbdoc: Get an array of method names for a given service
	End Rem
	Method GetServiceMethods:String[](serviceName:String) Abstract {rmiblocking}
		
	Rem
	bbdoc: Get a method signature as "(TYPENAME,TYPENAME,..)" e.g. "(String,Object,Int)"
	End Rem
	Method GetMethodSignature:String(serviceName:String , methodName:String) Abstract {rmiblocking}
End Type
'*** PROXY ***
Include "TRMIServiceRegistry-proxy.bmx"

Rem
bbdoc: Invoke a remote method
about: This function calls a method via fscom.rmi protocol using the stream and the service name given as @context and the given parameters. 
It is compatible with CreateProxy of fscom.proxy and is normally called by objets returned by RMIGetServiceObject in 
fscom.rmiclient or TRMIClient.GetRemoteServiceObject() in fscom.rmiserver.

The @context parameter is expected to be an #TRMIInvokeRemoteContext.

This method will wait #RMI_BLOCKING_TIMEOUT milliseconds for the remote method's return/throw if @metadata contains "rmiblocking"
and will not wait for any reaction of the remote method if @metadata contains "rminonblocking".
However if none is included in @metadata it will consider #RMI_DEFAULT_METHOD_BLOCKING.

If blocking, the method will submit any incoming invocations to the executor. Then it will return the value returned by the remote method or
throw the object thrown by the remote method. Also it will throw a #TRMIException if a connection problem occurs. 
Remote methods are unable to throw them via network, so if you can catch such exceptions to detect connection problems.
End Rem
Function RMIInvokeRemote:Object(serviceObject:Object , methodName:String , params:Object[] , metadata:String[], context:Object)
	Local ctx:TRMIInvokeRemoteContext = TRMIInvokeRemoteContext(context)
	If ctx = Null
		Throw TRMIException.Create("Illegal context for RMIInvokeRemote")
	EndIf
	
	Local block:Byte=MethodBlocking(metadata:String)
	Local channel:Byte
	If block
		channel=TRMIChannel.Alloc()
	Else 
		channel=255
	EndIf
	
	?threaded
		ctx.streamMutex.Lock()
	?
	Try 
		ctx.stream.WriteByte channel
		ctx.stream.WriteByte Asc("i")
		
		WriteBBString ctx.stream,ctx.serviceName
		WriteBBString ctx.stream,methodName
		
		For Local param:Object = EachIn params
			StreamSerializeObject param,ctx.stream
		Next
	
		ctx.stream.Flush()
	Catch ex:Object
	
		?threaded
			ctx.streamMutex.Unlock()
		?
	
		Throw TRMIException.Create("Unable to send invocation",ex)
	End Try
	?threaded
		ctx.streamMutex.Unlock()
	?

	
	If block
		Local time:Int=MilliSecs()+RMI_BLOCKING_TIMEOUT
		Repeat
			Select ctx.stream.Eof()
			Case 1
				?threaded
					ctx.streamMutex.Unlock()
				?

				Throw New TRMIException.Create("Connection lost")
			Case 0
				RMIPollStream(ctx)
			Default
				Delay 1
			End Select
			
			Local ans:TRMIChannel=TRMIChannel.GetAnswer(channel)
			If ans
				If ans.exception
					Throw ans.exception
				EndIf
				Return ans.retVal
			EndIf
		Until MilliSecs() > time
		Throw TRMIException.Create("Timeout")
	EndIf 
End Function

Type TRMIInvokeRemoteContext Extends TRMIContext
	Field serviceName:String	
End Type

Rem
If you want to manage an RMI server by yourself, use this method. 
Create a #TRMIContext object to provide this function with all information.
End Rem
Function RMIPollStream(ctx:TRMIContext)
	?threaded
		ctx.streamMutex.Lock()
	?
	Local answerChannel:Byte = ctx.stream.ReadByte()
	Local answerCmd:Byte = ctx.stream.ReadByte()
	Select answerCmd
	Case Asc("i")
		Local task:TTask = ReadInvocation(ctx , answerChannel)
		?threaded
			ctx.streamMutex.Unlock()
		?
		ctx.executor.Submit(task)
	Case Asc("r")
		Local retVal:Object = StreamDeserializeObject(ctx.stream)
		?threaded
			ctx.streamMutex.Unlock()
		?
		TRMIChannel.SetAnswer(answerChannel,retVal,Null)
	Case Asc("t")
		Local exception:Object = StreamDeserializeObject(ctx.stream)
		?threaded
			ctx.streamMutex.Unlock()
		?
		TRMIChannel.SetAnswer(answerChannel,Null,exception)
	Default
		Throw TRMIException.Create("Protocol error")
	End Select
End Function

Type TRMIContext
	Field stream:TStream
	?threaded
		'This mutex must be passed on with the stream so only one thread is able to use the stream
		'Only unlock the mutex if you have written a full rmi package.
		Field streamMutex:TMutex  
	?
	'stuff for incoming invocations
	Field serviceSuffix:String 'Only used by TypeServices
	Field executor:TExecutor
End Type

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'Package 'This stuff is for internal use only

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Type TRMIService Abstract
	Function GetServiceByName:TRMIService(name:String)
		Return TRMIService(serviceNameMap.ValueForKey(name))
	End Function
	
	Method Register()
		serviceNameMap.Insert(GetName() , Self)	
	End Method
	
	Method Unregister()
		serviceNameMap.Remove(GetName())		
	End Method
	Method GetName:String() Abstract
	Method GetTypeId:TTypeId() Abstract

	'Sort rmi services by their names
	Method Compare:Int(with:Object)
		Local with2:TRMIService = TRMIService(with)
		If with2
			Return GetName().Compare(with2.GetName())
		Else
			Return Super.Compare(with)
		EndIf
	End Method
End Type

Type TRMIObjectService Extends TRMIService
	Function GetServiceByObject:TRMIObjectService(serviceObject:Object)
		Return TRMIObjectService(TRMIObjectServiceImpl.objectMap.ValueForKey(serviceObject))
	End Function
	
	Method GetServiceObject:Object() Abstract	
End Type


Global RMI_POLL_NETWORK_HOOK:Int=AllocHookId()

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Global serviceNameMap:TMap = New TMap


Type TRMIObjectServiceImpl Extends TRMIObjectService
	Global objectMap:TMap = New TMap
	
	Field serviceObject:Object
	Field name:String
	Field typeid:TTypeId
	
	Method GetName:String()
		Return name
	End Method
	
	Method GetServiceObject:Object()
		Return serviceObject
	End Method
	
	Method GetTypeId:TTypeId()
		Return typeid
	End Method
	
	Method Register()
		objectMap.Insert(serviceObject , Self)
		Super.Register()
	End Method
	
	Method Unregister()
		objectMap.Remove(serviceObject)
		Super.Unregister()
	End Method
End Type

Type TRMIServiceRegistryImpl Extends TRMIServiceRegistry
	Const BUFFER_SIZE:Short=1024
	Global instance:TRMIServiceRegistry
	
	Method GetServices:String[]()
		Local services:String[BUFFER_SIZE]
		Local i:Int=0
		For Local service:String = EachIn serviceNameMap.Keys()
			Local at:Int=service.Find(RMI_PRIVATE_FLAG)
			
			If at<0
				If i = services.length
					services = services[..services.length + BUFFER_SIZE]
				EndIf
				services[i] = service
				i:+ 1
			EndIf
		Next
		services = services [..i]
		Return services
	End Method
	
	Method GetServiceMethods:String[](serviceName:String)
		'Private services are flaged with an @
		If serviceName.Find(RMI_PRIVATE_FLAG) >=0 Then Return New String[0]
		
		Local service:TRMIService = TRMIService.GetServiceByName(serviceName)
		If Not service Then Return Null
		
		Local methods:String[BUFFER_SIZE]
		Local i:Int=0
		For Local method_:TMethod=EachIn service.GetTypeId().Methods()
			If i = methods.length
				methods = methods[..methods.length + BUFFER_SIZE]
			EndIf
			methods[i] = method_.Name()
			i:+1
		Next
		methods = methods[..i]
		Return methods
	End Method
	
	Method GetMethodSignature:String(serviceName:String , methodName:String)
		'Private services are flaged with an @
		If serviceName.Find(RMI_PRIVATE_FLAG) >=0 Then Return ""

		Local service:TRMIService=TRMIService.GetServiceByName(serviceName)
		If Not service Then Return Null
		Local method_:TMethod = service.GetTypeId().FindMethod(methodName)
		If Not method_ Then Return Null
		
		Local signature:String = "("
		
		For Local paramType:TTypeId=EachIn method_.ArgTypes()
			signature:+ paramType.Name()+","
		Next
		signature = signature[..signature.length - 2]
		signature:+ ")"
		Return signature
	End Method
End Type

Type TRMIExceptionObjectFactory Extends TObjectFactory
	Global TRMIExceptionType:TTypeId=TTypeId.ForName("TRMIException")
	
	Method Serialize:Byte(obj:Object , typeid:TTypeId , stream:TStream , koTree:TKnownObjectTree)
		If typeId = TRMIExceptionType
			Return True
		EndIf
	End Method

	Method Deserialize:Object(typeid:TTypeId , stream:TStream, koTree:TKnownObjectTree)
		If typeId = TRMIExceptionType
			Return "TRMIException"
		EndIf
	End Method
End Type
New TRMIExceptionObjectFactory

Function MethodBlocking:Byte(metadata:String[])
	For Local meta:String = EachIn metadata
		Select meta
		Case "rminonblocking"
			Return False
		Case "rmiblocking"
			Return True
		End Select
	Next 
	Return RMI_DEFAULT_METHOD_BLOCKING
End Function

Function ReadInvocation:TTask(ctx:TRMIContext,channel:Byte)
	Local service:TRMIObjectService
	Local method_:TMethod
	Local params:Object[]
	Local serviceName:String
	Local methodName:String
	Try 
		'Get service
		serviceName = ReadBBString(ctx.stream)
		methodName = ReadBBString(ctx.stream)
		
		service = TRMIObjectService(TRMIService.GetServiceByName(serviceName))
		If Not service
			serviceName:+ ctx.serviceSuffix
			service = TRMIObjectService(TRMIService.GetServiceByName(serviceName))
			If Not service Then Throw "Service not found"
		EndIf
		
		'Get method
		method_ = service.GetTypeId().FindMethod(methodName)
		If Not method_ Then Throw "Method not found"
		
		'Get Params
		Local paramTypes:TTypeId[] = method_.ArgTypes()
		params = New Object[ paramTypes.length ]
		
		For Local i:Int = 0 Until paramTypes.length
			params[i]=StreamDeserializeObject(ctx.stream)
		Next
	Catch ex:Object
		Throw TRMIException.Create("Unable to read invocation",ex)
	End Try 
	
	Local obj:Object = service.GetServiceObject()
	
	If channel = 255
		Return TMethodTask.CreateR(obj , method_ , params)
	Else
		Local task:TRMITask = New TRMITask
		task.method_ = method_
		task.this = obj
		task.params = params
		
		task.ctx = ctx
		task.channel = channel
		Return task
	EndIf
End Function

Type TRMITask Extends TMethodTask
	Field channel:Byte
	Field ctx:TRMIContext
	
	'Override TMethodTask's Execute to not only invoke the method but
	'also send an answer
	Method Execute:Object()
		Try
			Local ret:Object = Super.Execute() 
			
			Try
				?threaded
					ctx.streamMutex.Lock()
				?
				ctx.stream.WriteByte channel
				ctx.stream.WriteByte Asc("r")
				StreamSerializeObject ret , ctx.stream
				ctx.stream.Flush()
			Catch ex:Object
				?threaded
					ctx.streamMutex.Unlock()
				?
				Throw TRMIException.Create("Unable to send answer" , ex)
			End Try
			?threaded
				ctx.streamMutex.Unlock()
			?
		Catch ex:TRMIException
			Throw ex
		Catch ex:Object
			Try
				?threaded
					ctx.streamMutex.Lock()
				?
				ctx.stream.WriteByte channel
				ctx.stream.WriteByte Asc("t")
				StreamSerializeObject ex,ctx.stream
				ctx.stream.Flush()
			Catch ex:Object
				?threaded
					ctx.streamMutex.Unlock()
				?
				Throw TRMIException.Create("Unable to send answer" , ex)
			End Try
			?threaded
				ctx.streamMutex.Unlock()
			?
		End Try
	End Method	
End Type

Type TRMIChannel
	Const STATE_PENDING:Int=0, STATE_ANSWERED:Int=1
	Global channels:TRMIChannel[254]
	'channel 255 can't be allocated: no answer for this channel
	?threaded
		Global mutex:TMutex = TMutex.Create()
	?

	Function Alloc:Byte()
		?threaded
			mutex.Lock()
		?
		For Local i:Int = 0 Until channels.length
			If channels[i] = Null
				channels[i] = New TRMIChannel
				?threaded
					mutex.Unlock()
				?
				Return i
			EndIf
		Next
		?threaded
			mutex.Unlock()
		?
		Throw TRMIException.Create("No free channels")
	End Function
	
	Function GetAnswer:TRMIChannel(i:Byte)
		If i = 255
			Throw TRMIException.Create("No answer for channel 255")
		EndIf
		?threaded
			mutex.Lock()
		?
		Local chan:TRMIChannel = channels[i]
		If chan = Null
			?threaded
				mutex.Unlock()
			?
			Throw TRMIException.Create("Channel " + i + " not set")
		EndIf
		
		If chan.state=STATE_PENDING
			?threaded
				mutex.Unlock
			?
			Return Null
		Else		
			channels[i] = Null
			?threaded
				mutex.Unlock
			?
			Return chan
		EndIf
	End Function
	
	Function SetAnswer(i:Byte,retVal:Object,exception:Object)
		If i = 255
			Throw TRMIException.Create("No answer for channel 255")
		EndIf
		?threaded
			mutex.Lock()
		?
		If channels[i] = Null
			?threaded
				mutex.Unlock()
			?
			Throw TRMIException.Create("Channel " + i + " not set")
		EndIf
		channels[i].retVal = retVal
		channels[i].exception = exception
		channels[i].state=STATE_ANSWERED
		?threaded
			mutex.Unlock
		?
	End Function

	Field retVal:Object
	Field exception:Object
	Field state:Int
End Type
