SuperStrict

Rem 
bbdoc: Serialization / Deserialization
about: Serialization means converting an Object to an aray of bytes to restore it from later.
Deserialization is the process of restoring a stored object.

This implementation of Serialization writes the byte array directly to a stream. 
Deserialization reads the byte array from a given stream.
End Rem
Module fscom.serialization

ModuleInfo "Version: 1.2"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

ModuleInfo "History: 1.2"
ModuleInfo "History: Fixed behaviour for empty Arrays (Null[])"
ModuleInfo "History: Switched from TTypeId.Fields to TTypeId.EnumFields to cover extended types"

Import brl.stream
Import brl.reflection

Rem
bbdoc: Serialize an object into a stream
End Rem
Function StreamSerializeObject(obj:Object , stream:TStream, knownObjectTree:TKnownObjectTree=Null)
	If obj = Null  
		WriteBBString stream , "Null"
		Return
	EndIf

	Local typeId:TTypeId = TTypeId.ForObject(obj)
	'Special case for empty arrays
	If typeId = ArrayTypeId
		WriteBBString stream , "Null[]"
		Return
	EndIf	

	WriteBBString stream,typeId.Name()
	
	If Not knownObjectTree Then knownObjectTree=New TKnownObjectTree
	Local fac:TObjectFactory=TObjectFactory.list
	Repeat
		If fac.Serialize(obj , typeId , stream, knownObjectTree)
			Return
		EndIf
		
		fac = fac._succ 
	Until fac = Null
	Throw "Unable to serialize object"
End Function

Rem
bbdoc: Deserialize an object from a stream
End Rem
Function StreamDeserializeObject:Object(stream:TStream, knownObjectTree:TKnownObjectTree=Null)
	Local typeName:String = ReadBBString(stream)
	
	'Null cases
	Select typeName
	Case "Null"
		Return Null
	Case "Null[]"
		'Empty array: Just returning null will result in array with length -1 if used with TMethod.Invoke
		Return New Object[0]
	End Select 

	If Not knownObjectTree Then knownObjectTree=New TKnownObjectTree
	
	Local typeId:TTypeId = TTypeId.ForName(typeName)
	If Not typeId Then Throw "Unknown Type " + typeName
	Local fac:TObjectFactory=TObjectFactory.list
	Repeat
		Local obj:Object=fac.Deserialize(typeId , stream, knownObjectTree)
		If obj
			Return obj
		EndIf
		
		fac = fac._succ 
	Until fac = Null
	Throw "Unable to deserialize object"
End Function

Rem
bbdoc: ObjectFactory base type
about: Extend this type to write serialization/deserialization methods for your custom objects. 
Then create an object of your type. #SerializeObject  and #DeserializeObject will try the newest
factory first, then go for older ones ending in a generic object factory that will try to serialize
your object by information from the reflection module.
End Rem
Type TObjectFactory
	Global list:TObjectFactory
	
	Field _succ:TObjectFactory
	Method New()
		_succ = list
		list = Self
	End Method
	
	Rem
	bbdoc: Try to serialize an object
	returns: True if successfully serialized an object. False if it's not a siutable serializer.
	about: Serialize will inspect @typeId and @obj to determine if it is able to
	serialize the object. If so, it will write the neccesary data into the stream.
	It will not touch the stream if it doesn't actually want to serialize the object.
	If an error occours an error should be thrown. Use @knownObjectTree to resolve cyclic 
	structures by searching for @obj in the known objects and only serialize it if it is 
	not found. In both cases insert the id returned by the methods of the known object tree.
	End Rem
	Method Serialize:Byte(obj:Object , typeId:TTypeId , stream:TStream, 	knownObjectTree:TKnownObjectTree) Abstract

	Rem
	bbdoc: Try to deserialize an object
	returns: The deserialized object or Null if it's not a siutable serializer
	about: Deserialize will inspect @typeId to determine if it is able to
	deserialize this kind of object. If so, it will read the data from the stream and build an object.
	It will not touch the stream to inspect the object and then decide to not to deserialize it.
	It can of course try to deserialize it and throw an error if the object was not serialized as expected. Use @knownObjectTree to resolve cyclic 
	structures by searching for a stored id in the known objects and only deserialize an object if the is 
	not found. Else just use the object returned by the known object tree.
	End Rem
	Method Deserialize:Object(typeId:TTypeId, stream:TStream, knownObjectTree:TKnownObjectTree) Abstract
End Type

Rem
bbdoc: Write a BlitzBasic style String to a stream: An integer for the length followed by the actual string
End Rem
Function WriteBBString(stream:TStream,str:String)
	stream.WriteInt str.length 
	stream.WriteString str
End Function

Rem
bbdoc: Read a BlitzBasic style String from a stream: An integer for the length followed by the actual string
End Rem
Function ReadBBString:String(stream:TStream)
	Local length:Int=stream.ReadInt()
	Return stream.ReadString(length)
End Function

'TODO: extends/implements a map-like structure to find objects faster
Type TKnownObjectTree
	Const BUFFER:Int=128
	Field array:Object[BUFFER]
	Field size:Int=0
	
	Method Find:Int(obj:Object)
		For Local i:Int = 0 Until size
			If obj = array[i]
				Return i
			EndIf
		Next
		Return -1
	End Method
	
	Method Insert:Int(obj:Object)
		If size => array.length
			array = array[..size + BUFFER]
		EndIf
		array[size] = obj
		size:+ 1
		Return size - 1
	End Method
	
	Method Get:Object(id:Int)
		If id < 0 Or id >= size Then Return Null
		Return array[id]
	End Method
End Type

Private

New TGenericObjectFactory
New TStringFactory
New TArrayFactory

Type TArrayFactory Extends TObjectFactory
	Method Serialize:Byte(obj:Object, typeId:TTypeId, stream:TStream, koTree:TKnownObjectTree)
		If typeId.ExtendsType(ArrayTypeId)
			Local id:Int = koTree.Find(obj)
			If id < 0
				id = koTree.Insert(obj)
				stream.WriteInt(id)
				Local dim:Byte=typeId.ArrayDimensions(obj)
				If dim>1
					Throw "Not yet implemented"
				EndIf 
				stream.WriteByte dim
				'For Local i:Byte = 0 Until dim
				Local length:Int = typeid.ArrayLength(obj , 0)
				stream.WriteInt length
				'Next
				
				For Local i:Int = 0 Until length'typeid.ArrayLength(obj,0)
					StreamSerializeObject( typeid.GetArrayElement(obj , i) , stream , koTree)
				Next
			Else
				stream.WriteInt(id)
			EndIf

			Return True
		EndIf
	End Method 
	
	Method Deserialize:Object(typeId:TTypeId, stream:TStream, koTree:TKnownObjectTree)
		If typeId.ExtendsType(ArrayTypeId)
			Local id:Int = stream.ReadInt()
			Local arr:Object = koTree.Get(id)
			If arr
				Return arr
			Else
				If id<>koTree.size Throw "Serialize / Deserialize KnownObjectTree out of sync"
			EndIf
			
			Local elementTypeId:TTypeId = typeId.ElementType()
			Local elementTypeName:String = elementTypeId.Name()
			Local dim:Byte = stream.ReadByte()
			If dim>1
				Throw "Not yet implemented"
			EndIf
			Local length:Int = stream.ReadInt()
	
			'We need a special case because NewArray will create an Object-Array
			'that won't cast to String[]
			If elementTypeName = "String"
				arr = New String[length]
			Else
				arr = typeid.NewArray(length)
			EndIf
	
			koTree.Insert(arr)
			For Local i:Int = 0 Until length
				typeid.SetArrayElement arr , i , StreamDeserializeObject(stream, koTree)
			Next
			
			Return arr
		EndIf
	End Method
End Type

Type TStringFactory Extends TObjectFactory
	Method Serialize:Byte(obj:Object, typeId:TTypeId, stream:TStream, koTree:TKnownObjectTree)
		If typeId = StringTypeId
			WriteBBString(stream,String(obj) )
			Return True
		EndIf
	End Method
	
	Method Deserialize:Object(typeId:TTypeId , stream:TStream, koTree:TKnownObjectTree)
		If typeId = StringTypeId
			Return ReadBBString(stream)
		EndIf
	End Method 
End Type

Type TGenericObjectFactory Extends TObjectFactory
	Method Serialize:Byte(obj:Object , typeId:TTypeId , stream:TStream, koTree:TKnownObjectTree)
		Local id:Int = koTree.Find(obj)
		If id < 0
			id = koTree.Insert(obj)
			stream.WriteInt(id)
			For Local f:TField = EachIn typeId.EnumFields()
				StreamSerializeObject f.Get(obj), stream , koTree
			Next
		Else
			stream.WriteInt(id)			
		EndIf
		Return True
	End Method

	Method Deserialize:Object(typeId:TTypeId , stream:TStream, koTree:TKnownObjectTree)
		Local id:Int = stream.ReadInt()
		Local obj:Object = koTree.Get(id)
		If obj
			Return obj
		Else
			If id<>koTree.size Throw "Serialize / Deserialize KnownObjectTree out of sync"
		EndIf

		obj = typeId.NewObject()
		koTree.Insert(obj)

		For Local f:TField = EachIn typeId.EnumFields()
			f.Set obj, StreamDeserializeObject(stream, koTree)
		Next

		Return obj
	End Method	
End Type
